/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * Copyright 2005-2006 Ambient Open Source Team
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: panelgroupclass.c,v 1.11 2009/09/27 19:21:55 kiero Exp $
 */

#include "ambient.h"

/* public */
#include <proto/graphics.h>
#include <proto/gadtools.h>

/* private */
#include "ambient_cat.h"
#include "mui_func.h"
#include "contextmenu.h"
#include "command.h"
#include "rexx.h"
#include "panelitem.h"
#include "smartreq.h"
#include "prefs.h"
#include "threads.h"
#include "gfx_dbuf.h"
#include "pointer.h"
#include "gfx_blit.h"
#include "playsound.h"

#define DIR_LEFTUP 0
#define DIR_RIGHTBOTTOM 1

struct Data {
	APTR configwin;
	ULONG size;
	ULONG horiz;
	ULONG dragmode;
	APTR cmenu;
	APTR dragobj[2];
	STRPTR framespec; /* string */
	STRPTR imagespec; /* string */
	ULONG childcnt;
	APTR lastobject;

	/* edit mode */
	ULONG locked;

	/* prefs saving */
	ULONG opened;

	/* dragging */

	ULONG draggable;

	/* zip mode */
	ULONG allowzipping;
	ULONG zipped;
	ULONG zipping;
	ULONG zipdir; /* DIR_LEFTUP or DIR_RIGHTBOTTOM */
	ULONG zipsnap_xs; /* original size of the panel */
	ULONG zipsnap_ys;
	struct doublebuf *dbuf;

	/* autozip */

	ULONG autozip, enable_autozip;
	struct MUI_InputHandlerNode ihnode;

	LONG oldmark;
	ULONG backdraw;

	/* menu stuff */
	struct MUI_EventHandlerNode ehnode;
	APTR childmenustrip;
	APTR childmenuobj;
	ULONG complex_menu;

	ULONG antibackfill;

	/* movemove stuff */
	ULONG movemode;
	APTR moveobj;
	ULONG movepos;
	ULONG *movearray;

	/* depth */
	ULONG depthmode;

	/* panel max width/height */

	ULONG max_w, max_h;

	/* orientation change flag for Refresh */
	ULONG oc_notify;
};

#define BD_DROPMARK (1 << 0L)


#define DOPANELLAYOUT(co,x,y,xs,ys) \
	({if (lm->lm_Type == MUILM_MINMAX) \
	{ \
		max_x = max(max_x, x + xs); \
		max_y = max(max_y, y + ys); \
	} \
	else \
	{ \
		if ((x + xs > lm->lm_Layout.Width) || (y + ys > lm->lm_Layout.Height) || !MUI_Layout(co, x, y, xs, ys, 0)) \
		{ \
			PDB(("layout failed\n")); \
			return (FALSE);\
		} \
	}})

#define DOPLACE(c) \
	({if (data->horiz) \
	{ \
		l = dist; \
		t = 0; \
	} \
	else \
	{ \
		l = 0; \
		t = dist; \
	}})


MUI_HOOK(layoutpanel, APTR grp, struct MUI_LayoutMsg *lm)
{
	struct Data *data = INST_DATA(OCLASS(grp), grp);

	switch (lm->lm_Type)
	{
		case MUILM_MINMAX:
			// fallthrough
		case MUILM_LAYOUT:
			{
				Object *cstate = (Object *)lm->lm_Children->mlh_Head;
				Object *child;
				ULONG type;
				ULONG t = 0, l = 0, w = 0, h = 0;
				ULONG max_x = 0, max_y = 0;
				ULONG childcount = 0;
				ULONG dist = 0;

				while ( (child = NextObject(&cstate)) )
				{
					if (_flags(child) & MADF_SHOWME)
					{
						type = 0;
						if (get(child, MA_Panel_Type, &type))
						{
							w = _minwidth(child);  /* XXX: remember only _minwidth() is known at that point. we should check min/max and give the size value of our panel */
							h = _minheight(child);

							switch (type)
							{
								case MV_Panel_Type_Drag:
									{
										/*
										 * Those gadgets always come
										 * first and/or last
										 */
										DOPLACE(child);
									}
									break;

								case MV_Panel_Type_Spacer:
								case MV_Panel_Type_Separator:
									{
										DOPLACE(child);
									}
									break;

								case MV_Panel_Type_Button:
								case MV_Panel_Type_ViewWatcher:
								case MV_Panel_Type_Bookmarks:
									{
										DOPLACE(child);
									}
									break;

#if USE_TRAY_LIB
								case MV_Panel_Type_TrayPanel:
									{
										DOPLACE(child);
									}
									break;
#endif

								case 0:
									{
										DOPLACE(child);
									}
									break;

								#ifdef DEBUG
								default:
									PDB(("no layout implemented for type %ld\n", type));
									break;
								#endif
							}

							DOPANELLAYOUT(child, l, t, w, h);

							childcount++;

							if (data->horiz)
							{
								dist += w;
							}
							else
							{
								dist += h;
							}
						}
#if 0
						#ifdef DEBUG
						else
						{
							PDB(("object %p <%s> has no type attribute!\n", child, OCLASS(child)->cl_ID));
						}
						#endif
#endif
					}
				}

				if (lm->lm_Type == MUILM_MINMAX)
				{
					struct Screen *scr = _screen(grp);

					/*
					 * Make sure the object is big enough for
					 * the cases where it's empty or almost
					 * empty. Also add the bottom/right frame
					 * space.
					 */
					if (data->horiz)
					{
						max_x = max(max_x, 12);
						max_y = max(max_y, data->size);
					}
					else
					{
						max_x = max(max_x, data->size);
						max_y = max(max_y, 12);
					}

					/*
					 * If we don't fit, set our size to the screen's size,
					 * then we'll fail in layout and use MUIM_AdjustLayout
					 * until we fit.
					 */
					max_x = min(max_x, scr->Width); /* XXX: frame size taken into account? wtf is that mess anyway */
					max_y = min(max_y, scr->Height);

					lm->lm_MinMax.MinWidth = max_x;
					lm->lm_MinMax.MinHeight = max_y;

					lm->lm_MinMax.DefWidth = max_x;
					lm->lm_MinMax.DefHeight = max_y;

					if (data->zipping || data->zipped)
					{
						lm->lm_MinMax.MaxWidth = MUI_MAXMAX;
						lm->lm_MinMax.MaxHeight = MUI_MAXMAX;
					}
					else
					{
						lm->lm_MinMax.MaxWidth = max_x;
						lm->lm_MinMax.MaxHeight = max_y;
					}

					data->max_w = max_x;
					data->max_h = max_y;
				}

				#if 0
				if (lm->lm_Type == MUILM_MINMAX)
				{
					dprintf("final childs: %ld\n", childcount);
					dprintf("MinWidth: %ld, MinHeight: %ld\nDefWidth: %ld, DefHeight: %ld\nMaxWidth: %ld, MaxHeight: %ld\n",
						(LONG)lm->lm_MinMax.MinWidth, (LONG)lm->lm_MinMax.MinHeight,
						(LONG)lm->lm_MinMax.DefWidth, (LONG)lm->lm_MinMax.DefHeight,
						(LONG)lm->lm_MinMax.MaxWidth, (LONG)lm->lm_MinMax.MaxHeight
					);
				}
				else
				{
					dprintf("layouting..\n");
				}
				#endif

			}
			return (TRUE);
	}
	return (MUILM_UNKNOWN);
}


DEFMMETHOD(AskMinMax)
{
	GETDATA;
	DOSUPER;

	/* XXX: beware.. we don't use += here because MUI adds too much frames atm.. experiment later */

	if (!data->zipping && !data->zipped)
	{
		struct MUI_MinMax *mm = msg->MinMaxInfo;
		LONG w, h;

		w = _screen(obj)->Width;
		h = _screen(obj)->Height;

		if (mm->MinWidth > w || mm->DefWidth > w || mm->MaxWidth > w)
		{
			mm->MinWidth = w;
			mm->DefWidth = w;
			mm->MaxWidth = w;

		}
		else if (mm->MinHeight > h || mm->DefHeight > h || mm->MaxHeight > h)
		{
			mm->MinHeight = h;
			mm->DefHeight = h;
			mm->MaxHeight = h;
		}
	}

	return (0);
}


static VOID start_autozip_timer(APTR obj, struct Data *data)
{
	if (muiRenderInfo(obj) == NULL || _app(obj) == NULL )
		return;

	if (data->ihnode.ihn_Object != NULL)
	{
		DoMethod(_app(obj), MUIM_Application_RemInputHandler, &data->ihnode);
	}

	/* 5 seconds */

	memset(&data->ihnode, 0, sizeof(struct MUI_InputHandlerNode));
	data->ihnode.ihn_Millis = 5000;
	data->ihnode.ihn_Object = obj;
	data->ihnode.ihn_Flags = MUIIHNF_TIMER;
	data->ihnode.ihn_Method = MM_Panelgroup_DelayedZip;

	DoMethod(_app(obj), MUIM_Application_AddInputHandler, &data->ihnode);
}


static VOID run_autozip_trigger(APTR obj, struct Data *data)
{
	/* XXX: run autozip timer. we should check if mouse is still above panel.. bah.. */

	if (!data->zipped && !data->zipping && data->enable_autozip)
	{
		start_autozip_timer(obj, data);
	}
}


#define RFR_CHANGED  (1 << 0UL)
#define RFR_REFRESH  (1 << 1UL)
#define RFR_REFRESH2 (1 << 2UL)

static void doset(APTR obj, struct Data *data, struct TagItem *tags, ULONG init)
{
	ULONG rfr = 0;

	FORTAG(tags)
	{
		case MA_Panelgroup_Configwin:
			data->configwin = (APTR)tag->ti_Data;
			break;
	
		case MA_Panelgroup_Size:
			data->size = tag->ti_Data;
			rfr |= (RFR_CHANGED | RFR_REFRESH);
			break;

		case MA_Panelgroup_Depth:
			data->depthmode = tag->ti_Data;
			rfr |= RFR_CHANGED;
			break;

		case MA_Panelgroup_Horiz:
			data->horiz = tag->ti_Data;
			data->oc_notify = TRUE;
            rfr |= (RFR_CHANGED | RFR_REFRESH);
			break;

		case MA_Panelgroup_DragMode:
			data->draggable = tag->ti_Data;
			data->enable_autozip = data->allowzipping && data->autozip && data->draggable;
			DoMethod(obj, MM_Panelgroup_ChangeDrag, tag->ti_Data);
			rfr |= RFR_CHANGED;
			break;
	
		case MA_Panelgroup_Zipping:
			data->allowzipping = tag->ti_Data;
			data->enable_autozip = data->allowzipping && data->autozip && data->draggable;
			rfr |= RFR_CHANGED;
			break;

		case MA_Panelgroup_AutoZip:
			data->autozip = tag->ti_Data;
			data->enable_autozip = data->allowzipping && data->autozip && data->draggable;
			if (!init)
			{
				run_autozip_trigger(obj, data);
			}
			break;

		case MA_Panelgroup_FrameSpec:
			if (tag->ti_Data)
			{
				if (data->framespec)
				{
					free(data->framespec);
				}
				if ( (data->framespec = malloc(strlen((STRPTR)tag->ti_Data) + 1)) )
				{
					strcpy(data->framespec, (STRPTR)tag->ti_Data);
				}
				/* XXX */
			}
			else
			{
				PDB(("no framespec.. wtf\n"));
			}
			rfr |= RFR_REFRESH2;
			break;
		
		case MA_Panelgroup_ImageSpec:
			if (tag->ti_Data)
			{
				if (data->imagespec)
				{
					free(data->imagespec);
				}
				if ( (data->imagespec = malloc(strlen((STRPTR)tag->ti_Data) + 1)) )
				{
					strcpy(data->imagespec, (STRPTR)tag->ti_Data);

					SetAttrs(obj,
						MUIA_Background,    NULL,            /* hmm??? */
						MUIA_Background,    data->imagespec,
						MUIA_Group_Forward, FALSE,
					TAG_DONE);
				}
				/* XXX */
			}
			break;

		case MA_Panelgroup_Locked:
			data->locked = tag->ti_Data;
			rfr |= RFR_CHANGED;
			break;
	}
	NEXTTAG

	if (rfr & (RFR_REFRESH | RFR_REFRESH2))
	{
		DoMethod(app, MUIM_Application_PushMethod,
			obj, 2, MM_Panelgroup_Refresh, rfr & RFR_REFRESH2
		);
	}

	if (rfr & RFR_CHANGED)
	{
		set(obj, MA_Panelgroup_HasChanged, TRUE);

		/* also restart autozipping trigger */
		run_autozip_trigger(obj, data);
	}
}


/* Returns panel object (for example tray panel)
 * which is parent (not always immediate parent)
 * to specified object.
*/      
static APTR get_panel_object(APTR obj)
{
	ULONG type;

	while (obj && !get(obj, MA_Panel_Type, &type))
		obj = _parent(obj);

	return obj;
}

DEFNEW
{
	struct Data *data;

	obj = DoSuperNew(cl, obj,
		MUIA_Group_LayoutHook, (ULONG)&layoutpanel_hook,
		MUIA_ContextMenu, 1,
	End;

	if (!obj)
	{
		return (NULL);
	}

	data = INST_DATA(cl, obj);
	data->oldmark = -1;
	data->cmenu = NULL;

	/* not really necessary but <shrug>, see panelwinclass for defaults */
	if ( (data->framespec = malloc(7)) )
	{
		strcpy(data->framespec, "000000");
	}
	/* XXX */

	if ( (data->imagespec = malloc(5)) )
	{
		strcpy(data->imagespec, "2:m2");
	}
	/* XXX */

	data->ihnode.ihn_Object = NULL;

	doset(obj, data, INITTAGS, TRUE);

	return ((ULONG)obj);
}


DEFDISP
{
	GETDATA;

	if (data->dbuf)
	{
		gfx_dbuf_free(data->dbuf);
	}

	if (data->framespec)
	{
		free(data->framespec);
	}

	if (data->imagespec)
	{
		free(data->imagespec);
	}

	if (data->ihnode.ihn_Object != NULL)
	{
		DoMethod(_app(obj), MUIM_Application_RemInputHandler, &data->ihnode);
	}

	return (DOSUPER);
}


DEFMMETHOD(Setup)
{
	ULONG rc;
	GETDATA;

	if ((rc = DOSUPER) && _win(obj))
	{
		data->ehnode.ehn_Object = obj;
		data->ehnode.ehn_Class = cl;
		data->ehnode.ehn_Events = IDCMP_MOUSEBUTTONS | IDCMP_INACTIVEWINDOW | IDCMP_MOUSEOBJECT;
		data->ehnode.ehn_Priority = 1; /* XXX: ok ? */
		data->ehnode.ehn_Flags = MUI_EHF_GUIMODE;
		DoMethod(_win(obj), MUIM_Window_AddEventHandler, (ULONG)&data->ehnode);

		run_autozip_trigger(obj, data);
	}
	return (rc);
}


DEFSMETHOD(Panelgroup_MoveMode_Start)
{
	GETDATA;

	if (!data->locked)
	{
		ULONG x, y;

		if (msg->mode == MV_Panelgroup_MoveMode_Start_Standalone)
		{
			ASSERT(data->childmenuobj);

			x = _left(data->childmenuobj);
			y = _top(data->childmenuobj);
			data->moveobj = get_panel_object(data->childmenuobj);
		}
		else
		{
			x = msg->x;
			y = msg->y;
			data->moveobj = (APTR)getv(_win(obj), MUIA_Window_MouseObject);
		}
		data->movepos = DoMethod(obj, MM_Panelgroup_WhichPosition, x, y);

		
		if (data->movepos != MV_Panelgroup_WhichPosition_None)
		{
			data->movemode = TRUE;

			/*
			 * Create a backup list.
			 */
			ASSERT(!data->movearray);

			if ( (data->movearray = malloc(data->childcnt * sizeof(ULONG) + sizeof(ULONG) + sizeof(ULONG))) ) /* + NULL terminator and MethodID */
			{
				ULONG *p = data->movearray;

				*p++ = MUIM_Group_Sort;

				FORCHILD(obj, MUIA_Group_ChildList)
				{
					*p++ = (ULONG)child;
				}
				NEXTCHILD
				*p = NULL;
			}

			/*
			 * Hm, for some weird reasons MUI gets confused if I don't
			 * disable then reenable the eventhandler. Probably my
			 * contextmenu stuff.
			 */
			DoMethod(_win(obj), MUIM_Window_RemEventHandler, (ULONG)&data->ehnode);
			data->ehnode.ehn_Events |= IDCMP_MOUSEMOVE;
			DoMethod(_win(obj), MUIM_Window_AddEventHandler, (ULONG)&data->ehnode);

			ASSERT(data->moveobj);
		}
		pointer_set(_window(obj), POINTER_MOVE);
	}
	return (0);
}


DEFSMETHOD(Panelgroup_MoveMode_Stop)
{
	GETDATA;

	ASSERT(data->movemode);

	data->ehnode.ehn_Events &= ~IDCMP_MOUSEMOVE;
	data->movemode = FALSE;

	pointer_clear(_window(obj));

	if (msg->mode == MV_Panelgroup_MoveMode_Stop_Abort)
	{
		if (data->movearray)
		{
			DoMethod(obj, MUIM_Group_InitChange);
			DoMethodA(obj, (Msg)&data->movearray[0]);
			DoMethod(obj, MUIM_Group_ExitChange);

		}
	}
	else
	{
		set(obj, MA_Panelgroup_HasChanged, TRUE);
	}
			
	if (data->movearray)
	{
		free(data->movearray);
		data->movearray = NULL;
	}
	return (0);
}


DEFMMETHOD(HandleEvent)
{
	GETDATA;

	if (msg->imsg)
	{
		switch (msg->imsg->Class)
		{
			case IDCMP_MOUSEOBJECT:
			{
				APTR o;

				o = (APTR)DoMethod(obj, MUIM_WhichObject, msg->imsg->MouseX, msg->imsg->MouseY);

				if (!o)
				{
					if (data->enable_autozip)
					{
						if (!data->zipped && !data->zipping)
						{
							start_autozip_timer(obj, data);
						}
					}
				}
				else if (data->ihnode.ihn_Object != NULL)
				{
					DoMethod(_app(obj), MUIM_Application_RemInputHandler, &data->ihnode);
					data->ihnode.ihn_Object = NULL;
				}

				if (!_conf(icon_hover))
					break;

				if (o == obj)
					o = NULL;

				if (o != data->lastobject)
				{
					if (data->lastobject)
					{
						set(data->lastobject, MA_Panel_Highlighted, FALSE);
					}

					data->lastobject = o;

					if (o)
					{
						set(data->lastobject, MA_Panel_Highlighted, TRUE);
					}
				}
			}
			break;

			case IDCMP_MOUSEBUTTONS:
			{
				switch (msg->imsg->Code)
				{

					case MENUDOWN:
					{
						GETDATA;
						APTR o;

						o = (APTR)getv(_win(obj), MUIA_Window_MouseObject);

						if (data->movemode)
						{
							DoMethod(obj, MM_Panelgroup_MoveMode_Stop, MV_Panelgroup_MoveMode_Stop_Abort);
							return (MUI_EventHandlerRC_Eat);
						}

						if (o)
						{
							data->childmenuobj = o;

						}
#if 0
						if (o && (o != obj))
						{
							ULONG type;

							data->complex_menu = TRUE;

							if (get(o, MA_Panel_Type, &type))
							{
								if (type == MV_Panel_Type_Drag)
								{
									data->complex_menu = FALSE;
								}
							}
							data->childmenuobj = o;
							data->childmenustrip = (APTR)DoMethod(o, MUIM_ContextMenuBuild, msg->imsg->MouseX, msg->imsg->MouseY);

							set(_win(obj), MUIA_Window_MouseObject, obj); /* we want the coming MUIM_ContextMenuBuild for us, no the child (XXX: should we restore it afterwards ?) */
						}
						else
						{
							data->childmenuobj = NULL;
							data->childmenustrip = NULL;
							data->complex_menu = FALSE;
						}
#endif

					}
					break;

					case MIDDLEDOWN:
					{
						DoMethod(obj, MM_Panelgroup_MoveMode_Start, MV_Panelgroup_MoveMode_Start_Coordinates, msg->imsg->MouseX, msg->imsg->MouseY);
					}
					break;

					case MIDDLEUP:
					{
						GETDATA;

						if (data->movemode)
						{
							DoMethod(obj, MM_Panelgroup_MoveMode_Stop, MV_Panelgroup_MoveMode_Stop_Keep);
						}
					}
					break;

					case SELECTDOWN:
					{
						GETDATA;

						if (data->movemode)
						{
							DoMethod(obj, MM_Panelgroup_MoveMode_Stop, MV_Panelgroup_MoveMode_Stop_Keep);

							return (MUI_EventHandlerRC_Eat);
						}
					}
				}
			}
			break;
		
			case IDCMP_MOUSEMOVE:
			{
				GETDATA;

				if (data->movemode)
				{
					ULONG newpos;

					newpos = DoMethod(obj, MM_Panelgroup_WhichPosition, msg->imsg->MouseX, msg->imsg->MouseY);

					if ((newpos != MV_Panelgroup_WhichPosition_None) && newpos != data->movepos )
					{
						APTR o;

						if ((o = (APTR)DoMethod(obj, MM_Panelgroup_WhichObject, msg->imsg->MouseX, msg->imsg->MouseY)) && (o != obj))
						{
							LONG doswap = FALSE;

							/*
							 * Exchange both objects is mouse passed half of the other object
							 */

							if ( data->horiz )
							{
								/* moving left */
								if ( newpos < data->movepos && (LONG)msg->imsg->MouseX < _left( o ) + _width( o ) / 2 )
								{
									doswap = TRUE;
								}
								/* moving right */
								if ( newpos > data->movepos && (LONG)msg->imsg->MouseX > _left( o ) + _width( o ) / 2 )
								{
									doswap = TRUE;
								}
							}
							else
							{
								/* moving up */
								if ( newpos < data->movepos && (LONG)msg->imsg->MouseY < _top( o ) + _height( o ) / 2 )
								{
									doswap = TRUE;
								}
								/* moving down */
								if ( newpos > data->movepos && (LONG)msg->imsg->MouseY > _top( o ) + _height( o ) / 2 )
								{
									doswap = TRUE;
								}
							}


							if ( doswap )
							{
								DoMethod(obj, MUIM_Group_InitChange);

								DoMethod(obj, MUIM_Group_MoveMember, o, data->movepos);
								DoMethod(obj, MUIM_Group_MoveMember, data->moveobj, newpos);

								DoMethod(obj, MUIM_Group_ExitChange);

								data->movepos = newpos;
							}
						}
					}
					return (MUI_EventHandlerRC_Eat); /* do not confuse plugins */
				}
			}
			break;

			case IDCMP_INACTIVEWINDOW:
			{
				GETDATA;

				if (data->movemode)
				{
					DoMethod(obj, MM_Panelgroup_MoveMode_Stop, MV_Panelgroup_MoveMode_Stop_Keep);
				}
			}
			break;
		}
	}
	return (0);
}


DEFMMETHOD(Cleanup)
{
	GETDATA;

	DoMethod(_win(obj), MUIM_Window_RemEventHandler, (ULONG)&data->ehnode);

	if (data->ihnode.ihn_Object != NULL)
		DoMethod(_app(obj), MUIM_Application_RemInputHandler, (ULONG)&data->ihnode);

	data->ihnode.ihn_Object = NULL;

	return (DOSUPER);
}

APTR menu_merge(APTR rootmenu, APTR menu);


APTR menu_merge(APTR rootmenu, APTR menu)
{
	return rootmenu;


}


DEFMMETHOD(ContextMenuBuild)
{
	GETDATA;

	if (!data->zipping)
	{
		ULONG flags = NULL;
#if 0
		if (data->cmenu)
		{
			MUI_DisposeObject(data->cmenu);
		}

		if (!data->locked)
		{
			flags |= AS_PANEL_UNLOCKED;
		}

		if (data->complex_menu)
		{
			ULONG attr;
			ASSERT(data->childmenuobj);

			if (get(data->childmenuobj, MA_Panel_Properties, &attr) && attr && !data->locked)
			{
				flags |= AS_PANEL_PROPERTIES;
			}
			if (get(data->childmenuobj, MA_Panel_About, &attr) && attr)
			{
				flags |= AS_PANEL_ABOUT;
			}
			if (get(data->childmenuobj, MA_Panel_Help, &attr) && attr)
			{
				flags |= AS_PANEL_HELP;
			}
		}
#endif

		if (data->childmenuobj)
		{

			APTR o;

			int i = 0;
			o = data->childmenuobj;

			while (((o = _parent(o)) && (o != obj)))
			{
				APTR menu;

				if (i++ > 10)
					break;

				menu = (APTR)DoMethod(o, MUIM_ContextMenuBuild, 0, 0);

				if (menu)
                {
					if (data->childmenustrip)
						DoMethod(data->childmenustrip, OM_ADDMEMBER, menu);
					else
						data->childmenustrip = menu;
				}
			}

			data->cmenu = data->childmenustrip;

			if (data->cmenu)
				DoMethod(data->cmenu, OM_ADDMEMBER, contextmenu_build(CM_PANELGROUP, 0));
			
            return ((ULONG)data->cmenu);
		}

#if 0
		if ( (data->cmenu = contextmenu_build(data->complex_menu ? CM_PANELGROUP_SUB : CM_PANELGROUP, flags)) ) /* XXX: hm perhaps find a trick to disable other menu entries if data->locked == TRUE */
		{

			/* set the proper mode */
			contextmenu_item_check(data->cmenu, "Panel ToggleLock", data->locked);
			
			if (data->childmenustrip)
			{
				STRPTR title;

				/*
				 * If there's a title, show it.
				 */
				if ( (title = (STRPTR)getv(data->childmenustrip, MUIA_Menu_Title)) )
				{
					set(data->cmenu, MUIA_Menu_Title, title);
				}
				return ((ULONG)data->childmenustrip);
				/* insert the childmenus, XXX: warning!! a childmenustrip can't be inserted directly! lot of processing here (title too) */
				/* XXX: do that.. then on ContextMenuChoice I'd have to iterate to get a count to know who I am, etc.. */
			}
			return ((ULONG)data->cmenu);
		}
		errormsg(ERR_NOMEM);
#endif
    }
	return (0);
}

DEFMMETHOD(ContextMenuAdd)
{
	ULONG flags = 0;
	GETDATA;

	if (!data->locked)
	{
		flags |= AS_PANEL_UNLOCKED;
	}

	data->cmenu	= contextmenu_build_simple(CM_PANELGROUP, flags);

	if (data->cmenu)
	{
		contextmenu_item_check(data->cmenu, "Panel ToggleLock", data->locked);
		DoMethod(msg->menustrip, OM_ADDMEMBER, data->cmenu);
	}
	return (DOSUPER);
}


DEFMMETHOD(ContextMenuChoice)
{
	GETDATA;
	ULONG rc = 0;
	struct command_menu *cm;

	if (data->cmenu && DoMethod(data->cmenu, MUIM_FindObject, msg->item))
	{
		if ( ((cm = (struct command_menu *)getv(msg->item, MA_Menuitem_Command))) )
		{
			if (cm->name && *cm->name)
				execute_command(_parent(obj), cm->type, cm->args, NULL);
		}
	}
	else
		rc = DOSUPER;
	
	data->cmenu = NULL;

	return (rc);
}


DEFGET
{
	switch (msg->opg_AttrID)
	{
		case MA_Panelgroup_Configwin:
		{
			GETDATA;
			*msg->opg_Storage = (ULONG)data->configwin;
			return (TRUE);
		}

		case MA_Panelgroup_Size:
		{
			GETDATA;
			*msg->opg_Storage = data->size;
			return (TRUE);
		}

		case MA_Panelgroup_Depth:
		{
			GETDATA;
			*msg->opg_Storage = data->depthmode;
			return (TRUE);
		}

		case MA_Panelgroup_Horiz:
		{
			GETDATA;
			*msg->opg_Storage = data->horiz;
			return (TRUE);
		}

		case MA_Panelgroup_DragMode:
		{
			GETDATA;
			*msg->opg_Storage = data->dragmode;
			return (TRUE);
		}
	
		case MA_Panelgroup_Zipping:
		{
			GETDATA;
			*msg->opg_Storage = data->allowzipping;
			return (TRUE);
		}

		case MA_Panelgroup_AutoZip:
		{
			GETDATA;
			*msg->opg_Storage = data->autozip;
			return (TRUE);
		}

		case MA_Panelgroup_FrameSpec:
		{
			GETDATA;
			*msg->opg_Storage = (ULONG)data->framespec;
			return (TRUE);
		}

		case MA_Panelgroup_ImageSpec:
		{
			GETDATA;
			*msg->opg_Storage = (ULONG)data->imagespec;
			return (TRUE);
		}

		case MA_Panelgroup_Locked:
		{
			GETDATA;
			*msg->opg_Storage = data->locked;
			return (TRUE);
		}

		case MA_Panelgroup_HasChanged:
			{
				/*
				 * Always TRUE just to trigger the notify.
				 */
				*msg->opg_Storage = TRUE;
				return (TRUE);
			}
	}
	return (DOSUPER);
}


DEFSET
{
	GETDATA;

	doset(obj, data, INITTAGS, FALSE);

	return (DOSUPER);
}


DEFADDMEMBER
{
	GETDATA;

	data->childcnt++;
	
	/* no MA_Panelgroup_HasChanged because changing the size does it anyway */

	return (DOSUPER);
}


DEFREMMEMBER
{
	GETDATA;

	data->childcnt--;

	/* ditto */

	return (DOSUPER);
}


DEFTMETHOD(Panelgroup_Delete)
{
	if (muiRenderInfo(obj) && _win(obj))
	{
		smartreq_request(_win(obj), _win(obj), GSI( MSG_PANELGROUPCLASS_TITLE), MM_Panelwin_Delete_Ok, 0, GSI( MSG_PANELGROUPCLASS_DELETECANCEL), MV_Notification_Warning, GSI( MSG_PANELGROUPCLASS_DELETEMSG), NULL); /* XXX: errmsg or so.. the displaybeep stuff */
	}
	/* XXX */
	return (0);
}


/*
 * If the layouthook failed, try to recover here.
 */
DEFMMETHOD(AdjustLayout)
{
	struct MinList *l;
	struct Node *on;

	if (msg->pass == 0)
	{
		return (TRUE); /* first pass doesn't require us to do anything (if that is even enabled in MUI) */
	}

	PDB(("adjustling layout..\n"));

	l = (struct MinList *)getv(obj, MUIA_Group_ChildList);

	/*
	 * Remove the last object, as long as it's not a
	 * paneldrag
	 */
	on = LASTNODE(l);

	if (getv(BASEOBJECT(on), MA_Panel_Type) == MV_Panel_Type_Drag)
	{
		on = PREVNODE(on);
	}

	if (!on || getv(BASEOBJECT(on), MA_Panel_Type) == MV_Panel_Type_Drag)
	{
		PDB(("holy fuck.. we are stuffed\n"));
		return (FALSE);
	}

	while (on && !(_flags(BASEOBJECT(on)) & MADF_SHOWME))
	{
		on = PREVNODE(on);
	}

	if (on)
	{
		_flags(BASEOBJECT(on)) &= ~MADF_SHOWME;
		PDB(("removed an object..\n"));
		
		return (TRUE); /* relayout */
	}
	return (FALSE);
}


/*
 * Send a Cleanup/Setup pair to all objects,
 * then redraw.
 */
DEFSMETHOD(Panelgroup_Refresh)
{
	GETDATA;

	if (msg->hard)
	{
		DoMethod(_parent(obj), MUIM_Group_InitChange);
	
		if (muiRenderInfo(obj) && _win(obj))
		{
			set(obj, MUIA_Frame, data->framespec);
			DoMethod(_win(obj), MM_Panelwin_AttachToBorder, FALSE);
		}
		else
		{
			set(obj, MUIA_Frame, data->framespec);
		}

		DoMethod(_parent(obj), MUIM_Group_ExitChange);
	}
	else
	{
		struct Window *window;

		if (muiRenderInfo(obj) && _win(obj))
		{
			window = _window(obj);


			DoMethod(obj, MUIM_Group_InitChange);

			data->oldmark = -1;

			DoMethod(obj, MUIM_Group_ExitChange);

			/*
			 * Readjust window size. Also reset zipped state
			 */
			window = _window(obj);
			data->zipped = FALSE;
			ChangeWindowBox(window, window->LeftEdge, window->TopEdge, window->MinWidth, window->MinHeight);
			DoMethod(_win(obj), MM_Panelwin_AttachToBorder, FALSE);
			run_autozip_trigger(obj, data);
		}
	}
	
	return (0);
}



DEFMMETHOD(Draw)
{
	GETDATA;

	if (data->zipping)
	{
		ASSERT(data->dbuf);

		switch (data->zipdir)
		{
			case DIR_LEFTUP:
				ASSERT(data->dragobj[0]);
				if (data->horiz)
				{
					gfx_blit(data->dbuf->bm, _rp(obj),
						BLITTAG_DstType, BLITVAL_DstType_RastPort,
						BLITTAG_SrcX, data->zipsnap_xs - _width(obj) + _width(data->dragobj[0]) + _left(data->dragobj[0]),
						BLITTAG_DstX, _width(data->dragobj[0]) + _left(data->dragobj[0]),
						BLITTAG_DstWidth, _width(obj),
					TAG_DONE);
				}
				else
				{
					gfx_blit(data->dbuf->bm, _rp(obj),
						BLITTAG_DstType, BLITVAL_DstType_RastPort,
						BLITTAG_SrcY, data->zipsnap_ys - _height(obj) + _height(data->dragobj[0]) + _top(data->dragobj[0]),
						BLITTAG_DstY, _height(data->dragobj[0]) + _top(data->dragobj[0]),
						BLITTAG_DstHeight, _height(obj),
					TAG_DONE);
				}
				break;

			case DIR_RIGHTBOTTOM:
				ASSERT(data->dragobj[1]);
				if (data->horiz)
				{
					ULONG rf = _right(obj) - _mright(obj);

					gfx_blit(data->dbuf->bm, _rp(obj),
						BLITTAG_DstType, BLITVAL_DstType_RastPort,
						BLITTAG_DstWidth, _width(obj) - _width(data->dragobj[1]),
					TAG_DONE);
					/* following makes flickering.. normal with the way a window resize works. we blit the right/bottom gadget */
					gfx_blit(data->dbuf->bm, _rp(obj),
						BLITTAG_DstType, BLITVAL_DstType_RastPort,
						BLITTAG_SrcX, data->zipsnap_xs - _width(data->dragobj[1]) - rf,
						BLITTAG_DstX, _width(obj) - _width(data->dragobj[1]) - rf,
						BLITTAG_DstWidth, _width(data->dragobj[1]) + rf,
					TAG_DONE);
				}
				else
				{
					ULONG bf = _bottom(obj) - _mbottom(obj);

					gfx_blit(data->dbuf->bm, _rp(obj),
						BLITTAG_DstType, BLITVAL_DstType_RastPort,
						BLITTAG_DstHeight, _height(obj) - _height(data->dragobj[1]),
					TAG_DONE);
					/* ditto */
					gfx_blit(data->dbuf->bm, _rp(obj),
						BLITTAG_DstType, BLITVAL_DstType_RastPort,
						BLITTAG_SrcY, data->zipsnap_ys - _height(data->dragobj[1]) - bf,
						BLITTAG_DstY, _height(obj) - _height(data->dragobj[1]) - bf,
						BLITTAG_DstHeight, _height(data->dragobj[1]) + bf,
					TAG_DONE);
				}
				break;
		}
	}
	else
	{
		/*
		 * This hack is because MUI clears by calling MUIM_Backfill
		 * on layout hooks. We skip the group's draw method so it
		 * doesn't do it.
		 */
		if (data->antibackfill)
		{
			data->antibackfill = FALSE;
			return (0);
		}
		
		DOSUPER;
	}
	return (0);
}


#if 0
DEFMMETHOD(Backfill) /* XXX: why not try to use MUIM_Draw instead ? except for a backfill/background fill perhaps.. the frame stuff should be in MUIM_Draw I think */
{
	GETDATA;

	if (data->backdraw & BD_DROPMARK)
	{
		SetAPen(_rp(obj), _pens(obj)[MPEN_FILL]);
		RectFill(_rp(obj), msg->left, msg->top, msg->right, msg->bottom);
	}
	else
	{
		SetAPen(_rp(obj), _pens(obj)[MPEN_BACKGROUND]); /* XXX: lame hack for now */
		RectFill(_rp(obj), msg->left, msg->top, msg->right, msg->bottom);
	}

	#if 0 /* XXX: need to find something nicer.. and configurable also */
	SetAPen(_rp(obj), _pens(obj)[MPEN_SHINE]);
	Move(_rp(obj), msg->right, msg->top);
	Draw(_rp(obj), msg->left, msg->top);
	Draw(_rp(obj), msg->left, msg->bottom);
	Move(_rp(obj), msg->left + 1, msg->bottom);

	SetAPen(_rp(obj), _pens(obj)[MPEN_SHADOW]);
	Draw(_rp(obj), msg->right, msg->bottom);
	Draw(_rp(obj), msg->right, msg->top);
	#endif

	return (TRUE); /* XXX: sure ? */
}
#endif

#define ADDDRAGOBJ(x) ({if (data->dragobj[x]) \
	{ \
		DoMethod(obj, OM_REMMEMBER, data->dragobj[x]); \
		MUI_DisposeObject(data->dragobj[x]); \
	} \
	if ( (data->dragobj[x] = NewObject(getpaneldragclass(), NULL, MA_Paneldrag_Type, x, TAG_DONE)) ) \
	{ \
		DoMethod(obj, OM_ADDMEMBER, data->dragobj[x]); \
		DoMethod(obj, MUIM_Group_MoveMember, data->dragobj[x], x ? -1 : 0); \
	}})

#define REMDRAGOBJ(x) ({if (data->dragobj[x]) \
	{ \
		DoMethod(obj, OM_REMMEMBER, data->dragobj[x]); \
		MUI_DisposeObject(data->dragobj[x]); \
		data->dragobj[x] = NULL; \
	}})

DEFSMETHOD(Panelgroup_ChangeDrag)
{
	GETDATA;

	DoMethod(obj, MUIM_Group_InitChange);

	data->dragmode = msg->dragmode;

	switch (data->dragmode)
	{
		case MV_Panelgroup_DragMode_None:
			REMDRAGOBJ(0);
			REMDRAGOBJ(1);
			break;

		case MV_Panelgroup_DragMode_LeftUp:
			ADDDRAGOBJ(0);
			REMDRAGOBJ(1);
			data->zipdir = DIR_LEFTUP;
			break;

		case MV_Panelgroup_DragMode_RightBottom:
			REMDRAGOBJ(0);
			ADDDRAGOBJ(1);
			data->zipdir = DIR_RIGHTBOTTOM;
			break;

		case MV_Panelgroup_DragMode_Both:
			ADDDRAGOBJ(0);
			ADDDRAGOBJ(1);
			break;
	
		#ifdef DEBUG
		default:
			PDB(("unimplemented dragmode: %ld\n", data->dragmode));
			break;
		#endif
	}
	
	set(obj, MA_Paneldrag_LockMode, data->locked); /* broadcast */

	DoMethod(obj, MUIM_Group_ExitChange);

	return (0);
}


DEFSMETHOD(Panelgroup_ToggleZip)
{
	GETDATA;

	if (data->zipping)
	{
		return (0);
	}

	switch (msg->type)
	{
		case MV_Paneldrag_Type_LeftUp:
			data->zipdir = DIR_LEFTUP;
			break;

		case MV_Paneldrag_Type_RightBottom:
			data->zipdir = DIR_RIGHTBOTTOM;
			break;

		#ifdef DEBUG
		default:
			PDB(("argh, wrong zipdir\n"));
			break;
		#endif
	}

	if (data->zipped)
	{
		if (!data->dbuf)
		{
			PDB(("unzipping from nothing NYI\n"));
			return (0);
		}

		data->zipping = TRUE;

		set(obj, MA_Paneldrag_Zipping, TRUE); /* broadcast */

		PLAYSFX(panelunzip);

		do_action(obj, TA_Panels_Zip,
			TT_Panels_Zip_Direction, FALSE,
			TT_Panels_Zip_Window, _window(obj),
			TT_Panels_Zip_XS, data->zipsnap_xs,
			TT_Panels_Zip_YS, data->zipsnap_ys,
			TT_Panels_Zip_Reversed, (data->zipdir == DIR_RIGHTBOTTOM) ? TRUE : FALSE,
			#if USE_PANEL_PREFS
			TT_Panels_Zip_ZipSpeed, _conf(panel_zipspeed),
			#else
			TT_Panels_Zip_ZipSpeed, MV_Panel_ZipSpeed_Medium,
			#endif
		TAG_DONE);
	}
	else
	{
		if (data->allowzipping)
		{
			data->zipsnap_xs = _width(obj);
			data->zipsnap_ys = _height(obj);

			ASSERT(_window(obj));

			if ( (data->dbuf = gfx_dbuf_alloc_reuse(data->dbuf, data->zipsnap_xs, data->zipsnap_ys, DBUF_DISPLAYABLE, _screen(obj)->RastPort.BitMap)) )
			{
				struct RastPort *oldrp;
				struct Window *oldwin;

				oldrp = _rp(obj);
				oldwin = _window(obj);
				
				_rp(obj) = data->dbuf->rp;
				_window(obj) = NULL;

				MUI_Redraw(obj, MADF_DRAWALL);

				_rp(obj) = oldrp;
				_window(obj) = oldwin;

				data->zipping = TRUE;

				/*
				 * Set all the objects (except paneldrags) to
				 * hidden.
				 */
				DoMethod(obj, MUIM_Group_InitChange);
				{
					FORCHILD(obj, MUIA_Group_ChildList)
					{
						if (getv(child, MA_Panel_Type) == MV_Panel_Type_Drag) continue;

						_flags(child) &= ~MADF_SHOWME;
					}
					NEXTCHILD
				}
				DoMethod(obj, MUIM_Group_ExitChange);

				set(obj, MA_Paneldrag_Zipping, TRUE); /* broadcast */

				PLAYSFX(panelzip);

				do_action(obj, TA_Panels_Zip,
					TT_Panels_Zip_Direction, TRUE,
					TT_Panels_Zip_Window, _window(obj),
					TT_Panels_Zip_XS, _minwidth(obj),
					TT_Panels_Zip_YS, _minheight(obj),
					TT_Panels_Zip_Reversed, (data->zipdir == DIR_RIGHTBOTTOM) ? TRUE : FALSE,
					#if USE_PANEL_PREFS
					TT_Panels_Zip_ZipSpeed, _conf(panel_zipspeed),
					#else
					TT_Panels_Zip_ZipSpeed, MV_Panel_ZipSpeed_Medium,
					#endif
				TAG_DONE);
			}
			/* XXX */
		}
	}
	return (0);
}


DEFSMETHOD(Thread_Finished)
{
	GETDATA;

	ASSERT(msg->action == TA_Panels_Zip);
	
	data->zipping = FALSE;
	
	if (msg->taglist && GetTagData(TT_Panels_Zip_Direction, FALSE, msg->taglist))
	{
		data->zipped = TRUE;

		/*
		 * Refresh the dragobjs otherwise they
		 * are covered by crap intuition did
		 * when resizing.
		 */
		if (data->dragobj[0])
		{
			MUI_Redraw(data->dragobj[0], MADF_DRAWALL);
		}
		if (data->dragobj[1])
		{
			MUI_Redraw(data->dragobj[1], MADF_DRAWALL);
		}

	}
	else
	{
		data->zipped = FALSE;
		/*
		 * Put back all objects as visible.
		 */
		data->antibackfill = TRUE;

		DoMethod(obj, MUIM_Group_InitChange);
		{
			FORCHILD(obj, MUIA_Group_ChildList)
			{
				_flags(child) |= MADF_SHOWME;
			}
			NEXTCHILD
		}
		DoMethod(obj, MUIM_Group_ExitChange); /* XXX: sigh.. it flickers */

		/* XXX: start autozip timer. crap. imagine something better. */

		run_autozip_trigger(obj, data);
	}

	set(obj, MA_Paneldrag_Zipping, FALSE); /* broadcast */

	return (0);
}


DEFSMETHOD(Panelgroup_Lock)
{
	SetAttrs(obj,
		MA_Paneldrag_LockMode, msg->lock, /* broadcast */
		MA_Panelgroup_Locked, msg->lock,
		TAG_DONE
	);

	return (0);
}


DEFTMETHOD(Panelgroup_ToggleLock)
{
	GETDATA;

	data->locked ^= 1;

	SetAttrs(obj,
		MA_Paneldrag_LockMode, data->locked, /* broadcast */
		MA_Panelgroup_Locked, data->locked,
		TAG_DONE
	);
	
	return (0);
}


DEFMMETHOD(DragQuery)
{
	GETDATA;

	if (!data->locked && !data->zipped)
	{
		ULONG val;
		if (get(msg->obj, MA_Panelitem_list_IsList, &val) || get(msg->obj, MA_Icon_Path, &val) ||
		   (get(msg->obj, MA_Panel_Type, &val) && ((val == MV_Panel_Type_Button)
#if USE_TRAY_LIB
            || (val == MV_Panel_Type_TrayPanel)
#endif
            )))
		{
			if (getv(msg->obj, MA_Icon_Type) != MV_Icon_Type_MyComputer &&
				getv(msg->obj, MA_Icon_Type) != MV_Icon_Type_AppIcon)
				return (MUIV_DragQuery_Accept);
		}
	}
	return (MUIV_DragQuery_Refuse);
}


/*
 * Given an x/y coordinate from the window (0/0 is the top/left),
 * returns the position of the object.
 */
DEFSMETHOD(Panelgroup_WhichPosition)
{
	ULONG i = MV_Panelgroup_WhichPosition_None;
	APTR o;
	
	if ((o = (APTR)DoMethod(obj, MM_Panelgroup_WhichObject, msg->x, msg->y)) && (o != obj))
	{
		i = 0;

		FORCHILD(obj, MUIA_Group_ChildList)
		{
			if (o == child)
			{
				break;
			}
			i++;
		}
		NEXTCHILD
	}
	return (i);
}

DEFSMETHOD(Panelgroup_WhichObject)
{
	APTR o;

	if ((o = (APTR)DoMethod(_win(obj), MUIM_WhichObject, msg->x, msg->y)))
		o = get_panel_object(o);

	return ((ULONG)o);
}

DEFMMETHOD(DragDrop)
{
	#ifdef DEBUG
	if (msg->obj == obj)
	{
		PDB(("huh, how did you do that?\n"));
	}
	else
	#endif
	{
		ULONG dummy;

		/* if source is an icon, add button and place this icon in it */

		if (getv(msg->obj, MA_DragDrop_Type) == MV_DragDrop_Type_Icon)
		{
			APTR o = NewObject(getpanelbuttonclass(), NULL, TAG_DONE);

			if (o != NULL) /* XXX: no clue about the distance here.. */
			{
				DoMethod(obj, MUIM_Group_InitChange);
				DoMethod(obj, OM_ADDMEMBER, o);
				DoMethod(obj, MUIM_Group_ExitChange);

				DoMethod(o, MM_Panelbutton_AddIcon, msg->obj);
			}
			return 0;
		}

		/* Panelitem_list */
		if (get(msg->obj, MA_Panelitem_list_IsList, &dummy))
		{
			LONG id = MUIV_List_NextSelected_Start;
			struct itemlist_entry *ie;
			ULONG i;
			ULONG changed = FALSE;

			for (i = 0; ; i++)
			{
				DoMethod(msg->obj, MUIM_List_NextSelected, &id);

				if (id == MUIV_List_NextSelected_End) break;

				DoMethod(msg->obj, MUIM_List_GetEntry, id, &ie);

				if (ie->type == PT_BUTTON || ie->type == PT_VIEWWATCHER || ie->type == PT_BOOKMARKS)
				{
					/* XXX: handle the other types later on */
					GETDATA;
					APTR o;

					if ( (o = NewObject(ie->cl, NULL, MA_Panel_Distance, data->horiz ? (msg->x - _window(obj)->LeftEdge) : (msg->y - _window(obj)->TopEdge), TAG_DONE)) )
					{
						ULONG i;

						DoMethod(obj, MUIM_Group_InitChange);
						DoMethod(obj, OM_ADDMEMBER, o);
							
						/*
						 * Try to find out where to insert the object.
						 */
						i = DoMethod(obj, MM_Panelgroup_WhichPosition, msg->x - max(0, _window(obj)->LeftEdge), msg->y - _window(obj)->TopEdge);
							
						if (i != MV_Panelgroup_WhichPosition_None)
						{
							if (!(data->dragobj[0] && !i)) /* dropping on the first dragobj should insert at the end */
							{
								DoMethod(obj, MUIM_Group_MoveMember, o, i);
							}
						}
							
						DoMethod(obj, MUIM_Group_ExitChange);
							
						changed	= TRUE;
					}
				}

				if (ie->type == PT_SPACER || ie->type == PT_SEPARATOR)
				{
					GETDATA;
					APTR o;

					if ( (o = NewObject(ie->cl, NULL, MA_Panel_Distance, data->horiz ? (msg->x - _window(obj)->LeftEdge) : (msg->y - _window(obj)->TopEdge), TAG_DONE)) )
					{
						ULONG i;

						DoMethod(obj, MUIM_Group_InitChange);
						DoMethod(obj, OM_ADDMEMBER, o);

						/*
						 * Try to find out where to insert the object.
						 */
						i = DoMethod(obj, MM_Panelgroup_WhichPosition, msg->x - max(0, _window(obj)->LeftEdge), msg->y - _window(obj)->TopEdge);

						if (i != MV_Panelgroup_WhichPosition_None)
						{
							if (!(data->dragobj[0] && !i)) /* dropping on the first dragobj should insert at the end */
							{
								DoMethod(obj, MUIM_Group_MoveMember, o, i);
							}
						}

						DoMethod(obj, MUIM_Group_ExitChange);

						changed	= TRUE;
					}
				}

#if USE_TRAY_LIB
				if (ie->type == PT_TRAY_PANEL)
				{
					/* XXX: handle the other types later on */
					GETDATA;
					APTR o;

					if ( (o = NewObject(ie->cl, NULL, MA_Panel_Distance, data->horiz ? (msg->x - _window(obj)->LeftEdge) : (msg->y - _window(obj)->TopEdge), TAG_DONE)) )
					{
						ULONG i;

						DoMethod(obj, MUIM_Group_InitChange);
						DoMethod(obj, OM_ADDMEMBER, o);

						/*
						 * Try to find out where to insert the object.
						 */
						i = DoMethod(obj, MM_Panelgroup_WhichPosition, msg->x - max(0, _window(obj)->LeftEdge), msg->y - _window(obj)->TopEdge);

						if (i != MV_Panelgroup_WhichPosition_None)
						{
							if (!(data->dragobj[0] && !i)) /* dropping on the first dragobj should insert at the end */
							{
								DoMethod(obj, MUIM_Group_MoveMember, o, i);
							}
						}

						DoMethod(obj, MUIM_Group_ExitChange);

						changed	= TRUE;
					}
				}
#endif
			}

			if (changed)
			{
				set(obj, MA_Panelgroup_HasChanged, TRUE);
			}
		}
	}
	return (0);
}


DEFMMETHOD(Group_ExitChange)
{
	GETDATA;

	/*
	 * Reposition the dragbars properly
	 * if they exists.
	 */
	if (data->dragobj[0])
	{
		DoMethod(obj, MUIM_Group_MoveMember, data->dragobj[0], 0);
	}

	if (data->dragobj[1])
	{
		DoMethod(obj, MUIM_Group_MoveMember, data->dragobj[1], -1);
	}

	/*
	 * Sets the visible flag on all childs to attempt a new
	 * layout.
	 */
	if (!data->zipping)
	{
	    FORCHILD(obj, MUIA_Group_ChildList)
		{
			_flags(child) |= MADF_SHOWME;

			/*
			 * if Orientation was changed, need to notify
			 * children groups like traypanel.
			 */
#if USE_TRAY_LIB
            if (data->oc_notify)
			{
				if (getv(child, MA_Panel_Type) == MV_Panel_Type_TrayPanel)
				{
					DoMethod(child, MUIM_Group_InitChange);
					DoMethod(child, MUIM_Group_ExitChange);
					data->oc_notify = FALSE;
				}
			}
#endif
        }
		NEXTCHILD
	}
	return (DOSUPER);
}


/*
 * Prevents areaclass from drawing a stupid frame around.
 */
#if 0
DEFMMETHOD(DragBegin)
{
	return (0);
}
#endif

#define _isinframe(x,y) (_between(_left(obj) + _mleft(obj),(x),_right(obj) - _mright(obj)) && _between(_top(obj) + _mtop(obj),(y),_bottom(obj) - _mbottom(obj)))

DEFMMETHOD(DragReport)
{
	if (msg->update)
	{
		MUI_Redraw(obj, MADF_DRAWOBJECT); /* XXX: I *think*.. could be drawupdate with drawmark also */
		return (MUIV_DragReport_Continue);
	}

	if (msg->obj != obj)
	{
		#if 0
		GETDATA;

		/*
		 * XXX: doing that on backfill is not so great. Lot
		 * of side effects. What we should do is do it on MUIM_Draw
		 * but in some special way (still don't know how)
		 */
		if (data->horiz)
		{
			ULONG t, b;

			t = _mtop(obj) + data->inner_top;
			b = _mbottom(obj) - data->inner_bottom;

			if (data->oldmark != -1)
			{
				DoMethod(obj, MUIM_Backfill,
					data->oldmark, t, data->oldmark + 1, b,
					data->oldmark, t,
					NULL
				);
			}

			if (_isinframe(msg->x, msg->y))
			{
				data->backdraw |= BD_DROPMARK;

				DoMethod(obj, MUIM_Backfill,
					msg->x, t, msg->x + 1, b,
					msg->x, t,
					NULL
				);
				data->backdraw &= ~BD_DROPMARK;
				data->oldmark = msg->x;
			}
			else
			{
				data->oldmark = -1;
			}
		}
		else
		{
			ULONG l, r;

			l = _mleft(obj) + data->inner_left;
			r = _mright(obj) - data->inner_right;

			if (data->oldmark != -1)
			{
				DoMethod(obj, MUIM_Backfill,
					l, data->oldmark, r, data->oldmark + 1,
					l, data->oldmark,
					NULL
				);
			}

			if (_isinframe(msg->x, msg->y))
			{
				data->backdraw |= BD_DROPMARK;

				DoMethod(obj, MUIM_Backfill,
					l, msg->y, r, msg->y + 1,
					l, msg->y,
					NULL
				);
				data->backdraw &= ~BD_DROPMARK;
				data->oldmark = msg->y;
			}
			else
			{
				data->oldmark = -1;
			}
		}
		#endif

	}
	return (MUIV_DragReport_Continue);
}


DEFSMETHOD(Panelgroup_RemoveItem)
{
	GETDATA;
	APTR o;

	if (msg->o == MV_Panelgroup_RemoveItem_Current)
	{
		o = get_panel_object(data->childmenuobj);
		ASSERT(o);
	}
	else
	{
		o = msg->o;
	}

	DoMethod(obj, MUIM_Group_InitChange);

	DoMethod(obj, OM_REMMEMBER, o);

	DoMethod(obj, MUIM_Group_ExitChange);

	data->lastobject = NULL;

	MUI_DisposeObject(o);

	set(obj, MA_Panelgroup_HasChanged, TRUE);

	return (0);
}


DEFSMETHOD(Panelgroup_SaveConfig)
{
	GETDATA;

	if (!data->zipped) /* XXX: actually we could save a zipped state as well but it might be tricky */
	{
		APTR pl, pi; /* list, item */
		ULONG index = 0;

		prefspool_item_remove(msg->pctx, NULL, DSI_LISTPOOL_PANEL); /* purge the whole prefs so that it doesn't mess when items are removed */

		pl = prefspool_item_add(msg->pctx, NULL, DSI_LISTPOOL_PANEL, NULL, NULL);

		if (pl)
		{
			FORCHILD(obj, MUIA_Group_ChildList)
			{
				if (!(pi = prefspool_item_get(msg->pctx, pl, index | DSF_LISTPOOL, NULL, NULL)))
				{
					pi = prefspool_item_add(msg->pctx, pl, index | DSF_LISTPOOL, NULL, NULL);
				}

				if (pi)
				{
					setprefslong_lp(msg->pctx, pi, DSI_LISTPOOL_PANEL_TYPE, getv(child, MA_Panel_Type)); /* XXX */
					DoMethod(child, MM_Panel_SaveConfig, msg->pctx, index); /* XXX: implement, also msg->pctx would be stored in the plugin's subclass then used by the plugin prefs calls */
				}
				/* XXX */
				index++;
			}
			NEXTCHILD
		}
	}
	return (0);
}


DEFTMETHOD(Panelgroup_DelayedZip)
{
	GETDATA;

	if (data->ihnode.ihn_Object != NULL)
	{
		DoMethod(_app(obj), MUIM_Application_RemInputHandler, &data->ihnode);
		data->ihnode.ihn_Object = NULL;
	}


	if (data->autozip && data->allowzipping) /* prefs may have changed when delaying... */
	{
		DoMethod(obj, MM_Panelgroup_ToggleZip, data->zipdir);
	}

	return (0);
}


BEGINMTABLE
DECNEW
DECDISP
DECGET
DECSET
DECADDMEMBER
DECREMMEMBER
DECMMETHOD(AskMinMax)
DECMMETHOD(Setup)
DECMMETHOD(Cleanup)
DECMMETHOD(Group_ExitChange)
DECSMETHOD(Panelgroup_MoveMode_Start)
DECSMETHOD(Panelgroup_MoveMode_Stop)
DECMMETHOD(HandleEvent)
//DECMMETHOD(ContextMenuBuild)
DECMMETHOD(ContextMenuAdd)
DECMMETHOD(ContextMenuChoice)
DECTMETHOD(Panelgroup_Delete)
DECMMETHOD(AdjustLayout)
DECSMETHOD(Panelgroup_Refresh)
DECMMETHOD(Draw)
DECSMETHOD(Panelgroup_ChangeDrag)
DECSMETHOD(Panelgroup_ToggleZip)
DECSMETHOD(Thread_Finished)
DECSMETHOD(Panelgroup_Lock)
DECTMETHOD(Panelgroup_ToggleLock)
DECSMETHOD(Panelgroup_WhichPosition)
DECSMETHOD(Panelgroup_WhichObject)
//DECMMETHOD(DragBegin)
DECMMETHOD(DragQuery)
DECMMETHOD(DragReport)
DECMMETHOD(DragDrop)
DECSMETHOD(Panelgroup_RemoveItem)
DECSMETHOD(Panelgroup_SaveConfig)
DECTMETHOD(Panelgroup_DelayedZip)
ENDMTABLE

DECSUBCLASS_NC(MUIC_Group, panelgroupclass)
