/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: panelitemwinclass.c,v 1.7 2008/05/10 17:35:30 kiero Exp $
 */

#include "ambient.h"

/* public */

/* private */
#include "ambient_cat.h"
#include "methodstack.h"
#include "mui_func.h"
#include "screen.h"
#include "panelitem.h"


struct Data {
	APTR lv_item;
	APTR txt_name;
	APTR txt_version;
	APTR txt_author;
	APTR ft_descr;
};


DEFNEW
{
	struct Data *data;
	APTR lv_item;
	APTR txt_name, txt_version, txt_author;
	APTR ft_descr;

	obj = DoSuperNew(cl, obj,
		MUIA_Window_Screen,      get_screen(),
		MUIA_Window_ScreenTitle, screentitle,
		MUIA_Window_Title,       GSI(MSG_PANELITEMWINCLASS_TITLE),
		MUIA_Window_ID,          MAKE_ID('A','M','P','I'),
		MUIA_Window_ShowIconify, FALSE,
		MUIA_Window_ShowPrefs,   FALSE,
		MUIA_Window_ShowJump,    FALSE,
		MUIA_Window_ShowAbout,   FALSE,
		MUIA_Window_NoMenus,     TRUE,
		MUIA_Window_Width,       MUIV_Window_Width_Screen(30),
		WindowContents, VGroup,

			Child, HGroup,

				Child, lv_item = NewObject(getpanelitem_listclass(), NULL,
					MUIA_CycleChain,           TRUE,
					MUIA_Weight,               30, /* XXX: hm hm */
					MUIA_Listview_DragType,    MUIV_Listview_DragType_Immediate,
					MUIA_Listview_MultiSelect, MUIV_Listview_MultiSelect_None,
					TAG_DONE),

				Child, VGroup,
	
					Child, ColGroup(2),

						Child, MUICreateLabel( MSG_PANELITEMWINCLASS_NAME, MUIO_Label_LeftAligned|MUIO_Label_SingleFrame),
						Child, txt_name = MUICreateTextNoFrame( MSG_PANELITEMWINCLASS_NAME, NULL ),

						Child, MUICreateLabel( MSG_PANELITEMWINCLASS_VERSION, MUIO_Label_LeftAligned|MUIO_Label_SingleFrame),
						Child, txt_version = MUICreateTextNoFrame( MSG_PANELITEMWINCLASS_VERSION, NULL ),

						Child, MUICreateLabel( MSG_PANELITEMWINCLASS_AUTHOR, MUIO_Label_LeftAligned|MUIO_Label_SingleFrame),
						Child, txt_author = MUICreateTextNoFrame( MSG_PANELITEMWINCLASS_AUTHOR, NULL ),
					End,
					
					Child, HGroup,
						GroupFrameT(GSI(MSG_PANELITEMWINCLASS_DESCGROUPTITLE)),
						
						Child, ft_descr = FloattextObject,
							TextFrame, MUIA_ShortHelp, GSI(MSG_PANELITEMWINCLASS_DESCGROUPTITLE_HELP),
						End,
					
					End,

				End,

			End,

			Child, TextObject,
				TextFrame,
				MUIA_Background,    MUII_TextBack,
				MUIA_Text_Contents, GSI(MSG_PANELITEMWINCLASS_DRAGINFO),
			End,

		End,
	End;

	if (!obj)
	{
		return (NULL);
	}

	data = INST_DATA(cl, obj);

	data->lv_item     = lv_item;
	data->txt_name    = txt_name;
	data->txt_version = txt_version;
	data->txt_author  = txt_author;
	data->ft_descr    = ft_descr;

	DoMethod(obj, MUIM_Notify, MUIA_Window_CloseRequest, TRUE,
		obj, 1, MM_Panelitemwin_Close
	);

	DoMethod(data->lv_item, MUIM_Notify, MUIA_List_Active, MUIV_EveryTime,
		obj, 2, MM_Panelitemwin_ListChange, MUIV_TriggerValue
	);

	return ((ULONG)obj);
}


DEFGET
{
	switch (msg->opg_AttrID)
	{
		case MA_Window_ID:
			*msg->opg_Storage = 0;
			return (TRUE);

		case MA_Window_Path:
			*msg->opg_Storage = MV_Window_Path_None;
			return (TRUE);

		case MA_Window_Type:
			*msg->opg_Storage = MV_Window_Type_Panelitem;
			return (TRUE);
	}

	return (DOSUPER);
}


DEFSMETHOD(Panelitemwin_ListChange)
{
	GETDATA;

	if (msg->val == MUIV_List_Active_Off)
	{
		set(data->txt_name,    MUIA_Text_Contents,  "");
		set(data->txt_version, MUIA_Text_Contents,  "");
		set(data->txt_author,  MUIA_Text_Contents,  "");
		set(data->ft_descr,    MUIA_Floattext_Text, "");
	}
	else
	{
		struct itemlist_entry *ie;

		DoMethod(data->lv_item, MUIM_List_GetEntry, msg->val, &ie);

		ASSERT(ie);

		set(data->txt_name,   MUIA_Text_Contents,  ie->name   ? ie->name   : (STRPTR) GSI(MSG_PANELITEMWINCLASS_UNKNOWN));
		set(data->txt_author, MUIA_Text_Contents,  ie->author ? ie->author : (STRPTR) GSI(MSG_PANELITEMWINCLASS_UNKNOWN));
		set(data->ft_descr,   MUIA_Floattext_Text, ie->descr  ? ie->descr  : (STRPTR) GSI(MSG_PANELITEMWINCLASS_NONE));
		DoMethod(data->txt_version, MUIM_SetAsString, MUIA_Text_Contents, "%lu.%lu\n", ie->version, ie->revision);
	}

	return (0);
}


DEFTMETHOD(Panelitemwin_Close)
{
	methodstack_push(_app(obj), 2, MM_Application_DisposeWindow, obj);

	return (0);
}


BEGINMTABLE
DECNEW
DECGET
DECSMETHOD(Panelitemwin_ListChange)
DECTMETHOD(Panelitemwin_Close)
ENDMTABLE

DECSUBCLASS_NC(MUIC_Window, panelitemwinclass)
