/*
 * Ambient - the ultimate desktop
 * ------------------------------
 * Copyright 2001-2005 by David Gerber <zapek@morphos.net>
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Id: prefswin_panelclass.c,v 1.5 2006/09/18 23:17:30 fab Exp $
 */

#include "ambient.h"

#if USE_PANEL_PREFS

/* public */

/* private */
#include "ambient_cat.h"
#include "mui_func.h"
#include "prefswin.h"
#include "prefsclone.h"

struct Data {
	APTR sl_speed;
};


/* XXX: should use a table.. or better, some general routine taking a table. panel has other setting using that conversion stuff */
static ULONG speed_to_slider(ULONG speed)
{
	switch (speed)
	{
		case MV_Panel_ZipSpeed_Slow:
			return (0);

		case MV_Panel_ZipSpeed_Medium:
			return (1);

		case MV_Panel_ZipSpeed_Fast:
			return (2);
		
		default:
			return (3);
	}
}


static ULONG slider_to_speed(ULONG active)
{
	switch (active)
	{
		case 0:
			return (MV_Panel_ZipSpeed_Slow);

		case 1:
			return (MV_Panel_ZipSpeed_Medium);

		case 2:
			return (MV_Panel_ZipSpeed_Fast);

		default:
			return (MV_Panel_ZipSpeed_Instant);
	}
}


DEFNEW
{
	struct Data *data;
	APTR sl_speed;

	obj = DoSuperNew(cl, obj,

		Child, VSpace(0),

		Child, HGroup,
			
			Child, HSpace(-1),
			
			Child, Label1("Zipping speed:"),
			Child, sl_speed = NewObject(getpanelsliderspeedclass(), NULL, TAG_DONE), /* XXX: ditto */
			
			Child, HSpace(-1),
		
		End,

		Child, VSpace(0),

	End;

	if (!obj)
	{
		return (0);
	}

	data = INST_DATA(cl, obj);
	data->sl_speed = sl_speed;

	DoMethod(obj, MUIM_MultiSet, MUIA_CycleChain, 1,
		sl_speed, NULL
	);

	SetAttrs(sl_speed,
		MUIA_Slider_Level, speed_to_slider(getprefslong(DSI_PANEL_ZIPSPEED)),
		MUIA_ShortHelp, GSI(MSG_PREFSWIN_PANEL_ZIPSPEED_HELP),
		TAG_DONE
	);

	return ((ULONG)obj);
}


DEFTMETHOD(Prefswin_Store)
{
	GETDATA;

	setprefslong(DSI_PANEL_ZIPSPEED, slider_to_speed(getv(data->sl_speed, MUIA_Slider_Level)));

	return (0);
}


DEFDISPOSE
{
	DoMethod(obj, MM_Prefswin_Store);

	return (DOSUPER);
}


BEGINMTABLE
DECNEW
DECTMETHOD(Prefswin_Store)
DECDISPOSE
ENDMTABLE

DECSUBCLASS_NC(MUIC_Group, prefswin_panelclass)

#endif

